#!/usr/bin/env python
# coding: utf-8

# # Introduction

# Welcome to the PyAT tutorial.
# Here we will go through some of the basic syntax of PyAT. Slowly evolving into more complex examples.
# The outline of the tutorial is as follows:
# >    1. Defining some simple elements and creating a lattice object
# >    2. Simple tracking, orbit correction, radiation flags
# >    3. Accessing and modifying ring elements
# >    4. Matching
# 
# 

# First we can import PyAT (and other useful packages) and see all accessible functions using dir

# In[1]:


import numpy as np
import copy
import matplotlib.pyplot as plt
import at
#dir(at)


# Many functions! But we cannot do anything because we do not have a lattice. Let's define some key parameters for a DBA lattice. Here I am simply taking the values from the at/machine_data/dba.m file

# I want to define a drift. But how do I know what to provide to the function? We can use 'help' or ? to tell us.

# All elements are found in the file:
# > at/pyat/at/lattice/elements.py
# 
# This can also be accessed by:

# In[2]:


dir(at.elements)


# In[3]:


#help(at.elements.Drift)
get_ipython().run_line_magic('pinfo', 'at.elements.Drift')


# We can see that we need to provide first a 'family_name' as a string, and a 'length' as a float. We can also use the dir function to see all available attributes that can be accessed from the drift.

# In[4]:


DR_24 = at.Drift('DR_24', 2.4)

# To see all of the available attributes
dir(DR_24)
# i.e. Length can be re-set by DR_24.Length = New_Length


# ## A note on pointers 

# Python uses pointers. When creating a lattice, this can cause some headaches. Observe the following simple examples:

# In[5]:


a = np.arange(10) #the array is assigned some memory
b = a #b is now a pointer to the location of the memory for a
b[5] = 100 #we modify b[5], but it also modifies a due to pointers
print(a)


# In[6]:


drift = at.Drift('Test',5) # Drift of length 5m
pointer_test = [drift, drift] # Here we make a list of 2 drifts of 5m each

# Now we can reduce the length of the first drift
pointer_test[0].Length = 2
print(pointer_test[0].Length, pointer_test[1].Length)


# 
# 
# As you can see, the drift variable was placed twice in the list. When you modify the length of one, you are in fact modifying all. This can cause issues if you predefine elements (drifts, quads, sexts) and then place them in multiple places in the lattice. One must be careful!

# We can perform similar actions to define the needed dipoles, quadrupoles, sextupoles and markers

# In[7]:


Nhalfcells = 25 #actually it is 50 half cells
K_QF1 = 0.38041 #Strength of the quadrupoles
K_QD2 = -0.2708
K_QD3 = -0.33319
K_QF4 = 0.4588

Ks = 0.1 # Strength of the sextupoles
L1 = 3

# get arc lattice structure
arc_half=[at.Drift('DR_01', L1*6/8),
          at.Quadrupole('QF1', L1/4, K_QF1),
          at.Drift('DR_02', L1*2/8),
          at.Quadrupole('QD2', L1/4, K_QD2),
          at.Drift('DR_03', L1*2/8),
          at.Bend('Bend', L1, 2*np.pi/(2*Nhalfcells)),          
          at.Drift('DR_04', L1/16),
          at.Quadrupole('QD3', L1/4, K_QD3),
          at.Drift('DR_05', L1/16),
          at.Sextupole('SD', L1/16, -Ks),
          at.Drift('DR_06', L1*3/16),
          at.Quadrupole('QF4', L1/4, K_QF4),
          at.Drift('DR_07', L1/16),
          at.Sextupole('SF', L1/4, Ks)
          ]


# We have made a list of elements that represent a half cell consisting of drifts, dipoles, quadrupoles and sextupoles. This is just a python list. We will now use this half, make a full cell (first half + reverse of the first half) and convert it into a lattice object and start to benefit greatly from what PyAT has to offer.

# In[8]:


arc_half_copy = copy.deepcopy(arc_half) #This is to avoid problems with pointers when copying the half arc
arc = arc_half + [at.Monitor('BPM_CellCenter')] + arc_half_copy[::-1] #We also add a monitor in the center for later on
arc = at.Lattice(arc, energy=3e9)
print('Periodicity:', arc.periodicity)


# As can be seen. We had to define the energy of the ring. The periodicity of the ring has been automatically set based on the total bending angle of the cell. We can do a quick check of the linear optics using the plot_beta function.

# In[9]:


import at.plot
arc.plot_beta()


# Lets now make a full ring of 25 cells.

# In[10]:


ring = []
for i in np.arange(int(Nhalfcells)):
    ring += copy.deepcopy(arc)
    ring += [at.Monitor('BPM_SS')]
    ring += [at.Corrector('Corrector', 0, [0, 0])]


# In[11]:


ring = at.Lattice(ring, energy=3e9)
ring.plot_beta() 
geom,radius,ax = ring.plot_geometry(centered=True)
get_ipython().run_line_magic('pinfo', 'ring.plot_geometry')


# Lets check to see if our ring has the correct total bending angle. For this we need to use some functions to access certain lattice elements.
# 
# Here we will use 'get_uint32_index'. For all of these kinds of functions, you can specify either a string 'BPM' or a string with a wildcard 'BPM*' or an element in order to select the correct indices. More detail will be provded later on how to access and modify lattice elements. But for now we can see:

# In[12]:


bpm_inds = at.get_uint32_index(ring,'BPM*')

bpm_inds = at.get_uint32_index(ring, at.Monitor)
print(len(bpm_inds))



# So what do we want to do? We want to look at every Bend element, and look at its 'BendingAngle' attribute, then sum them all. To achieve this, we can use 'get_value_refpts. This uses 'get_uint32_index' to get the appropriate indexes for the elements, then extracts the given Attribute Name.

# In[13]:


dipole_bending_angles = at.get_value_refpts(ring, at.elements.Bend, 'BendingAngle')
np.sum(dipole_bending_angles) -  2 * np.pi


# In[14]:


quad_strengths = at.get_value_refpts(ring, at.elements.Quadrupole, 'PolynomB', 1)
#This searches in ring, for all elements that match a quadrupole, then finds
#the PolynomB attribute, and takes the second index (python indexes start at 0)

quad_strengths


# Similarly, parameters can be set using 'set_value_refpts'. For example, lets increase all quadrupole strengths by 1%

# For the final part of this lattice building section. We must discuss something very important.
# 
# If you call 'set_value_refpts' with copy=False (default behaviour), the values that you set will be modified for the ring lattice.
# If you call it with copy=True, you function will return a copy of the original lattice with the modifications made, so the correct usage is

# In[15]:


new_ring = at.set_value_refpts(ring, at.elements.Quadrupole, 'PolynomB', quad_strengths*5, 1, copy=True)
print(np.sum(at.get_value_refpts(ring, at.Quadrupole, 'PolynomB', 1)))
print(np.sum(at.get_value_refpts(new_ring, at.Quadrupole, 'PolynomB', 1)))


# This question of in-place modification will show up in many different locations (tracking, radiation flags).

# We have one thing missing! An RF Cavity! Let's generate it and then place it in the straight section of the 4th cell. We can use the BPM in the center of the straight section to find the right index

# In[16]:


from at.constants import clight
dir(at.constants)


# In[17]:


#at.elements.RFCavity?
harmonic_number = 992
frf = harmonic_number * clight / ring.circumference
rfca = at.elements.RFCavity('RF_Cavity', 0, 5e6, frf-100, harmonic_number, ring.energy, PassMethod='IdentityPass') 
#From the help, we can see we need family_name, length, voltage, frequency, harmonic_number, energy
straight_section_inds = at.get_uint32_index(ring, 'BPM_SS')
ring.insert(straight_section_inds[2], rfca)


# In[18]:


straight_section_inds


# In[19]:


at.get_uint32_index(ring, 'RF*')


# There are many other handy functions for modifying the lattice in this way including a lot of the standard python list modifications. These will be covered in detail in the next notebook.
# 
# 

# Now we have a pretty realistic looking lattice. But at the moment the radiation is off and longitudinal motion is disabled. Lets first enable the 6D tracking and finish setting up the cavity.
# There are two functions for this:
# 
# > ring.enable_6d
# 
# > ring.disable_6d
# 
# When we ask to switch on 6D, the function goes through every single pass method in the lattice, and if an equivalent pass method exists that contains the ending '*RadPass' instead of '*Pass' then it updates the pass method. It also switches all RFCavity elements to have CavityPass. The information in the help of the function allows the user to specifically give pass methods for certain groups of elements. 
# 
# When ring.disable_6d is called, the pass methods ending in *RadPass* are switched to their *Pass* equivalents, and the RFCavity is switched to IdentityPass (i.e. nothing) if it has length 0, or DriftPass if it has a non-zero length. This effectively disables longitudinal tracking, which is necessary for the accurate computation of uncoupled linear optics.
# 

# In[20]:


ring.enable_6d()
print(np.unique(at.get_value_refpts(ring, np.arange(len(ring)), 'PassMethod')))
ring.disable_6d()
print(np.unique(at.get_value_refpts(ring, np.arange(len(ring)), 'PassMethod')))


# In[21]:


# cavity on, radiation on
ring.enable_6d()

#cavity on, radiation off
ring.disable_6d(cavity_pass='RFCavityPass')

#cavity off, radiation on
ring.enable_6d(cavity_pass='IdentityPass')

#cavity off, radiation off
ring.disable_6d()


# Now we can switch on 6D tracking and use some in built lattice functions to set the RF frequency and the cavity phase.

# In[22]:


ring.enable_6d()
freq_old = ring.get_rf_frequency()
ring.set_rf_frequency()
freq_new = ring.get_rf_frequency()
freq_old - freq_new


# In[23]:


print(freq_old/1e6, 'MHz')
print(freq_new/1e6, 'MHz')
print(ring.harmonic_number, ring.revolution_frequency*ring.harmonic_number/1e6)


# As you can see. our initial guess of 352 MHz was not good for the specified harmonic number. In this simple case set_rf_frequency has applied the RF frequency from harmonic_number*f0.
# 
# We can also set the timelag of the cavity. If the TimeLag of the cavity is 0, the there is no phase shift on the RF wave, and the beam will settle itself at a synchronous position where it recovers the energy loss per turn. We can use the function 'set_cavity_phase' to numerically compute the value needed to ensure the beam oscillation point is at ct=0.

# In[24]:


print(ring.get_rf_timelag())
ring.set_cavity_phase()
print(ring.get_rf_timelag())


# We now have a pretty functional lattice. We can now use the function 'envelope_parameters' to numerically compute some interesting and important parameters (emittances, synchrotron frequency, damping numbers)

# In[25]:


envelope = at.envelope_parameters(ring)
print(envelope)


# We may also use the class radiation_parameters, but there is one key difference when compared with envelope_parameters. Radiation_parameters uses the analytical formula rather than numerically solving. So the results may differ in some cases as not all pass methods are included.

# In[26]:


arc


# One last thing on this lattice. Let's check and set the tunes and chromaticities using the in built functions

# In[27]:


qx, qy, qs = at.get_tune(ring)
print(qx,qy,qs)
get_ipython().run_line_magic('pinfo', 'at.fit_tune')



# In[28]:


a = at.fit_tune(ring, 'QF*', 'QD*', [0.45,0.85], copy=True)
print(ring.get_tune())


# In[29]:


Qpx, Qpy, Qpz = at.get_chrom(ring)
print(Qpx, Qpy, Qpz)


# In[30]:


at.fit_chrom(ring, at.get_uint32_index(ring, 'SF*'), at.get_uint32_index(ring,'SD*'), [0,0])
ring.get_chrom()


# In[31]:


ring.save('./dba.mat', mat_key='RING')
arc.save('./arc.mat', mat_key='ARC')


# In[32]:


ring.plot_beta()


# In[ ]:




